# -*- coding: utf-8 -*-
# scinstr/version.py

__version__ = "0.4.6"

# 0.4.6  (18/07/2025): Fix SDM3065 driver communication:
#                      use PyVISA package in place of USBTMC.
# 0.4.5  (31/03/2025): Minor: add capability to tctrl/l350.py module.
# 0.4.4  (24/02/2025): Add web interface to handle Counter emulator.
# 0.4.3  (10/01/2025): Add web interface to handle DMM emulator.
#                      Correct bugs in vacuum subpackage.
# 0.4.2  (29/02/2024): Add driver for SDM3065A and some correction.
# 0.4.1  (29/02/2024): Add subpakage 'vnav, 'compressor' and 'phimeter'.
#                      Major update of tctrl subpakage. Also a lot of code
#                      review, bug corrections and code cleaning.
# 0.4.0  (07/04/2023): Separate script files, now in scinstr-bin package,
#                      from driver files.
# 0.3.0  (06/03/2020): - Divide package into subpackage.
#                      - Name of signals use snake case instead of camel case.
# 0.2.0  (13/12/2019): Add support of signal/slot facilities in pure python
#                      using signalslot package.
# 0.1.3  (12/10/2019): Update daq34972a.py to PyQt5.
# 0.1.2  (19/03/2019): Move to PyQt5
# 0.1.1  (14/03/2019): Lot of modifications...
# 0.1.0a (18/09/2018): Initial version
